@@ -10,7 +10,7 @@ static void cc_from_x(point_ptr p, element_t x)
 
     element_init(t, p->curve->field);
     p->inf_flag = 0;
-    element_mul(t, x, x);
+    element_square(t, x);
     element_add(t, t, cc->a);
     element_mul(t, t, x);
     element_add(t, t, cc->b);
@@ -29,7 +29,7 @@ static void cc_random(point_ptr p)
     p->inf_flag = 0;
     do {
 	element_random(p->x);
-	element_mul(t, p->x, p->x);
+	element_square(t, p->x);
 	element_add(t, t, cc->a);
 	element_mul(t, t, p->x);
 	element_add(t, t, cc->b);
@@ -54,7 +54,7 @@ static void cc_from_hash(point_ptr p, int len, void *data)
     p->inf_flag = 0;
     for(;;) {
 	element_from_hash(p->x, len, datacopy);
-	element_mul(t, p->x, p->x);
+	element_square(t, p->x);
 	element_add(t, t, cc->a);
 	element_mul(t, t, p->x);
 	element_add(t, t, cc->b);
@@ -89,16 +89,15 @@ static inline void cc_double_no_check(point_ptr r, point_ptr p)
     //same point: double them
 
     //lambda = (3x^2 + a) / 2y
-    element_set_si(e0, 3);
-    element_mul(lambda, p->x, p->x);
-    element_mul(lambda, lambda, e0);
+    element_square(lambda, p->x);
+    element_mul_si(lambda, lambda, 3);
     element_add(lambda, lambda, cc->a);
     element_add(e0, p->y, p->y);
     element_invert(e0, e0);
     element_mul(lambda, lambda, e0);
     //x1 = lambda^2 - 2x
     element_add(e1, p->x, p->x);
-    element_mul(e0, lambda, lambda);
+    element_square(e0, lambda);
     element_sub(e0, e0, e1);
     //y1 = (x - x1)lambda - y
     element_sub(e1, p->x, e0);
@@ -164,7 +163,7 @@ static void cc_add(point_ptr r, point_ptr p, point_ptr q)
 	element_sub(lambda, q->y, p->y);
 	element_mul(lambda, lambda, e0);
 	//x3 = lambda^2 - x1 - x2
-	element_mul(e0, lambda, lambda);
+	element_square(e0, lambda);
 	element_sub(e0, e0, p->x);
 	element_sub(e0, e0, q->x);
 	//y3 = (x1-x3)lambda - y1
@@ -456,7 +455,7 @@ static void sn_random(point_ptr p)
     do {
 	element_random(p->x);
 	if (element_is0(p->x)) continue;
-	element_mul(t, p->x, p->x);
+	element_square(t, p->x);
 	element_add(t, t, p->x);
 	element_mul(t, t, p->x);
     } while (!element_is_sqr(t));
@@ -475,8 +474,7 @@ static inline void sn_double_no_check(point_ptr r, point_ptr p)
     //same point: double them
 
     //lambda = (3x^2 + 2x) / 2y
-    element_set_si(e0, 3);
-    element_mul(lambda, e0, p->x);
+    element_mul_si(lambda, p->x, 3);
     element_set_si(e0, 2);
     element_add(lambda, lambda, e0);
     element_mul(lambda, lambda, p->x);
@@ -485,7 +483,7 @@ static inline void sn_double_no_check(point_ptr r, point_ptr p)
     element_mul(lambda, lambda, e0);
     //x1 = lambda^2 - 2x - 1
     element_add(e1, p->x, p->x);
-    element_mul(e0, lambda, lambda);
+    element_square(e0, lambda);
     element_sub(e0, e0, e1);
     element_set_si(e1, 1);
     element_sub(e0, e0, e1);
@@ -553,7 +551,7 @@ static void sn_add(point_ptr r, point_ptr p, point_ptr q)
 	element_sub(lambda, q->y, p->y);
 	element_mul(lambda, lambda, e0);
 	//x3 = lambda^2 - x1 - x2 - 1
-	element_mul(e0, lambda, lambda);
+	element_square(e0, lambda);
 	element_sub(e0, e0, p->x);
 	element_sub(e0, e0, q->x);
 	element_set1(e1);
@@ -577,6 +575,59 @@ static void sn_clear(curve_ptr c)
 {
 }
 
+//nonsingular points on sn curves map to finite field elements via
+//(x, y) --> (y + x)/(y - x)
+//and the reverse map is
+//a --> (4a/(a-1)^2, 4a(a+1)/(a-1)^3)
+void sn_point_to_field(element_t out, point_ptr P)
+{
+    element_t e0, e1;
+    if (P->inf_flag) {
+	element_set1(out);
+	return;
+    }
+    element_init(e0, out->field);
+    element_init(e1, out->field);
+    element_add(e0, P->y, P->x);
+    element_sub(e1, P->y, P->x);
+    element_invert(e1, e1);
+    element_mul(out, e0, e1);
+    element_clear(e0);
+    element_clear(e1);
+}
+
+void sn_field_to_point(point_ptr P, element_t in)
+{
+    element_t e0, e1, e2;
+
+    if (element_is1(in)) {
+	point_set_inf(P);
+	return;
+    }
+    element_init(e0, in->field);
+    element_init(e1, in->field);
+    element_init(e2, in->field);
+
+    element_set1(e1);
+    element_sub(e0, in, e1);
+    element_invert(e0, e0);
+
+    element_mul_si(e2, in, 4);
+
+    element_add(P->y, in, e1);
+
+    element_mul(e1, e0, e0);
+    element_mul(P->x, e1, e2);
+    element_mul(P->y, P->y, e2);
+    element_mul(P->y, P->y, e0);
+    element_mul(P->y, P->y, e1);
+    P->inf_flag = 0;
+
+    element_clear(e0);
+    element_clear(e1);
+    element_clear(e2);
+}
+
 void curve_init_singular_with_node(curve_ptr c, field_t field)
 {
     c->field = field;
