@@ -13,7 +13,7 @@ typedef struct darray_s darray_t[1];
 typedef struct darray_s *darray_ptr;
 
 void darray_init(darray_t a);
-darray_ptr darray_new();
+darray_ptr darray_new(void);
 void darray_append(darray_t a, void *p);
 void *darray_at(darray_t a, int i);
 void darray_clear(darray_t a);
