@@ -5,6 +5,8 @@ struct e_pairing_data_s {
     field_t Fq;
     curve_t Eq;
     mpz_t tateexp;
+    int exp2, exp1;
+    int sign1, sign0;
 };
 typedef struct e_pairing_data_s e_pairing_data_t[1];
 typedef struct e_pairing_data_s *e_pairing_data_ptr;
@@ -159,6 +161,188 @@ void e_param_inp_str(e_param_ptr p, FILE *stream)
     symtab_clear(tab);
 }
 
+static void e_miller(element_t res, point_t P,
+	element_ptr Qx, element_ptr Qy, e_pairing_data_ptr p)
+{
+    //collate divisions
+    int n;
+    element_t v, vd;
+    element_t v1, vd1;
+    point_t Z, Z1;
+    element_t a, b, c;
+    common_curve_ptr cc = P->curve->data;
+    element_t t0;
+    element_t e0, e1;
+    int i;
+
+    void do_vertical(element_t e, point_ptr A)
+    {
+	if (point_is_inf(A)) {
+	    return;
+	}
+	element_sub(e0, Qx, A->x);
+	element_mul(e, e, e0);
+    }
+
+    void do_tangent(element_t e, point_ptr A)
+    {
+	//a = -slope_tangent(A.x, A.y);
+	//b = 1;
+	//c = -(A.y + a * A.x);
+	//but we multiply by 2*A.y to avoid division
+
+	//a = -Ax * (Ax + Ax + Ax + twicea_2) - a_4;
+	//Common curves: a2 = 0 (and cc->a is a_4), so
+	//a = -(3 Ax^2 + cc->a)
+	//b = 2 * Ay
+	//c = -(2 Ay^2 + a Ax);
+	element_ptr Ax = A->x;
+	element_ptr Ay = A->y;
+
+	if (point_is_inf(A)) {
+	    return;
+	}
+
+	if (element_is0(Ay)) {
+	    //order 2
+	    do_vertical(e, A);
+	    return;
+	}
+
+	element_square(a, Ax);
+	element_mul_si(a, a, 3);
+	element_add(a, a, cc->a);
+	element_neg(a, a);
+
+	element_add(b, Ay, Ay);
+
+	element_mul(t0, b, Ay);
+	element_mul(c, a, Ax);
+	element_add(c, c, t0);
+	element_neg(c, c);
+
+	element_mul(e0, a, Qx);
+	element_mul(e1, b, Qy);
+	element_add(e0, e0, e1);
+	element_add(e0, e0, c);
+	element_mul(e, e, e0);
+    }
+
+    void do_line(element_ptr e, point_ptr A, point_ptr B)
+    {
+	//a = -(B.y - A.y) / (B.x - A.x);
+	//b = 1;
+	//c = -(A.y + a * A.x);
+	//but we'll multiply by B.x - A.x to avoid division
+
+	element_ptr Ax = A->x;
+	element_ptr Ay = A->y;
+	element_ptr Bx = B->x;
+	element_ptr By = B->y;
+
+	//we assume B is never O
+	if (point_is_inf(A)) {
+	    do_vertical(e, A);
+	    return;
+	}
+
+	if (!element_cmp(Ax, Bx)) {
+	    if (!element_cmp(Ay, By)) {
+		do_tangent(e, A);
+		return;
+	    }
+	    do_vertical(e, A);
+	    return;
+	}
+
+	element_sub(b, Bx, Ax);
+	element_sub(a, Ay, By);
+	element_mul(t0, b, Ay);
+	element_mul(c, a, Ax);
+	element_add(c, c, t0);
+	element_neg(c, c);
+
+	element_mul(e0, a, Qx);
+	element_mul(e1, b, Qy);
+	element_add(e0, e0, e1);
+	element_add(e0, e0, c);
+	element_mul(e, e, e0);
+    }
+
+    element_init(a, P->curve->field);
+    element_init(b, P->curve->field);
+    element_init(c, P->curve->field);
+    element_init(t0, P->curve->field);
+    element_init(e0, res->field);
+    element_init(e1, res->field);
+
+    element_init(v, res->field);
+    element_init(vd, res->field);
+    element_init(v1, res->field);
+    element_init(vd1, res->field);
+    point_init(Z, P->curve);
+    point_init(Z1, P->curve);
+
+    point_set(Z, P);
+
+    element_set1(v);
+    element_set1(vd);
+    element_set1(v1);
+    element_set1(vd1);
+
+    n = p->exp1;
+    for (i=0; i<n; i++) {
+	element_square(v, v);
+	element_square(vd, vd);
+	do_tangent(v, Z);
+	point_double(Z, Z);
+	do_vertical(vd, Z);
+    }
+    if (p->sign1 < 0) {
+	element_set(v1, vd);
+	element_set(vd1, v);
+	do_vertical(vd1, Z);
+	point_neg(Z1, Z);
+    } else {
+	element_set(v1, v);
+	element_set(vd1, vd);
+	point_set(Z1, Z);
+    }
+    n = p->exp2;
+    for (; i<n; i++) {
+	element_square(v, v);
+	element_square(vd, vd);
+	do_tangent(v, Z);
+	point_double(Z, Z);
+	do_vertical(vd, Z);
+    }
+    element_mul(v, v, v1);
+    element_mul(vd, vd, vd1);
+    do_line(v, Z, Z1);
+    point_add(Z, Z, Z1);
+    do_vertical(vd, Z);
+
+    if (p->sign0 > 0) {
+	do_vertical(v, P);
+    }
+
+    element_invert(vd, vd);
+    element_mul(res, v, vd);
+
+    element_clear(v);
+    element_clear(vd);
+    element_clear(v1);
+    element_clear(vd1);
+    point_clear(Z);
+    point_clear(Z1);
+    element_clear(a);
+    element_clear(b);
+    element_clear(c);
+    element_clear(t0);
+    element_clear(e0);
+    element_clear(e1);
+}
+
 //assumes P is in the base field, Q in some field extension
 static void cc_miller(element_t res, mpz_t q, point_t P,
 	element_ptr Qx, element_ptr Qy)
@@ -190,7 +374,7 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
 
 	//a = -Zx * (Zx + Zx + Zx + twicea_2) - a_4;
 	//Common curves: a2 = 0 (and cc->a is a_4), so
-	//a = -(Zx (Zx + Zx + Zx) + cc->a)
+	//a = -(3 Zx^2 + cc->a)
 	//b = 2 * Zy
 	//c = -(2 Zy^2 + a Zx);
 	element_ptr Zx = Z->x;
@@ -206,9 +390,8 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
 	    return;
 	}
 
-	element_add(a, Zx, Zx);
-	element_add(a, a, Zx);
-	element_mul(a, a, Zx);
+	element_square(a, Zx);
+	element_mul_si(a, a, 3);
 	element_add(a, a, cc->a);
 	element_neg(a, a);
 
@@ -285,8 +468,8 @@ static void cc_miller(element_t res, mpz_t q, point_t P,
     m = mpz_sizeinbase(q, 2) - 2;
 
     while(m >= 0) {
-	element_mul(v, v, v);
-	element_mul(vd, vd, vd);
+	element_square(v, v);
+	element_square(vd, vd);
 	do_tangent(v);
 	point_double(Z, Z);
 	do_vertical(vd);
@@ -324,8 +507,13 @@ static void e_pairing(element_ptr out, element_ptr in1, element_ptr in2,
     point_random(R);
     element_init(e0, out->field);
     point_add(QR, Q, R);
-    cc_miller(out, pairing->r, in1->data, QR->x, QR->y);
-    cc_miller(e0, pairing->r, in1->data, R->x, R->y);
+    if (0) {
+	cc_miller(out, pairing->r, in1->data, QR->x, QR->y);
+	cc_miller(e0, pairing->r, in1->data, R->x, R->y);
+    } else {
+	e_miller(out, in1->data, QR->x, QR->y, p);
+	e_miller(e0, in1->data, R->x, R->y, p);
+    }
     element_invert(e0, e0);
     element_mul(out, out, e0);
     element_pow(out, out, p->tateexp);
@@ -345,6 +533,10 @@ void pairing_init_e_param(pairing_t pairing, e_param_t param)
     pairing->is_almost_coddh = generic_is_almost_coddh;
 
     p =	pairing->data = malloc(sizeof(e_pairing_data_t));
+    p->exp2 = param->exp2;
+    p->exp1 = param->exp1;
+    p->sign1 = param->sign1;
+    p->sign0 = param->sign0;
     field_init_fp(p->Fq, param->q);
     element_init(a, p->Fq);
     element_init(b, p->Fq);
@@ -367,6 +559,7 @@ void pairing_init_e_param(pairing_t pairing, e_param_t param)
     element_clear(b);
 }
 
+//TODO: the following code is useless as the Tate pairing is degenerate on singular curves
 //assumes P is in the base field, Q in some field extension
 static void sn_miller(element_t res, mpz_t q, point_t P,
 	element_ptr Qx, element_ptr Qy)
@@ -411,8 +604,7 @@ static void sn_miller(element_t res, mpz_t q, point_t P,
 	    return;
 	}
 
-	element_add(a, Zx, Zx);
-	element_add(a, a, Zx);
+	element_mul_si(a, Zx, 3);
 	element_set_si(e0, 2);
 	element_add(a, a, e0);
 	element_mul(a, a, Zx);
@@ -545,7 +737,6 @@ void pairing_init_singular_with_node(pairing_t pairing, mpz_t q)
     mpz_t one;
 
     mpz_init(pairing->r);
-    //mpz_set(pairing->r, q);
     mpz_sub_ui(pairing->r, q, 1);
     field_init_fp(pairing->Zr, pairing->r);
     pairing->map = sn_pairing;
