@@ -33,6 +33,9 @@ struct field_s {
     void (*add)(element_ptr, element_ptr, element_ptr);
     void (*sub)(element_ptr, element_ptr, element_ptr);
     void (*mul)(element_ptr, element_ptr, element_ptr);
+    void (*mul_mpz)(element_ptr, element_ptr, mpz_ptr);
+    void (*mul_si)(element_ptr, element_ptr, signed long int);
+    void (*square)(element_ptr, element_ptr);
     void (*invert)(element_ptr, element_ptr);
     void (*neg)(element_ptr, element_ptr);
     int (*cmp)(element_ptr, element_ptr);
@@ -112,6 +115,21 @@ static inline void element_mul(element_ptr n, element_ptr a, element_ptr b)
     n->field->mul(n, a, b);
 }
 
+static inline void element_mul_mpz(element_ptr n, element_ptr a, mpz_ptr z)
+{
+    n->field->mul_mpz(n, a, z);
+}
+
+static inline void element_mul_si(element_ptr n, element_ptr a, signed long int z) 
+{
+    n->field->mul_si(n, a, z);
+}
+
+static inline void element_square(element_ptr n, element_ptr a)
+{
+    n->field->square(n, a);
+}
+
 static inline void element_neg(element_ptr n, element_ptr a)
 {
     n->field->neg(n, a);
@@ -196,9 +214,5 @@ static inline void field_clear(field_ptr f)
 
 element_ptr field_get_nqr(field_ptr f);
 
-static inline void field_init(field_ptr f)
-{
-    f->nqr = NULL;
-    mpz_init(f->order);
-}
+void field_init(field_ptr f);
 #endif //FIELD_H
