@@ -122,6 +122,45 @@ static void fq_mul(element_ptr n, element_ptr a, element_ptr b)
     element_clear(e2);
 }
 
+static void fq_mul_mpz(element_ptr n, element_ptr a, mpz_ptr z)
+{
+    fq_data_ptr p = a->data;
+    fq_data_ptr r = n->data;
+    element_mul_mpz(r->x, p->x, z);
+    element_mul_mpz(r->y, p->y, z);
+}
+
+static void fq_mul_si(element_ptr n, element_ptr a, signed long int z)
+{
+    fq_data_ptr p = a->data;
+    fq_data_ptr r = n->data;
+    element_mul_si(r->x, p->x, z);
+    element_mul_si(r->y, p->y, z);
+}
+
+static void fq_square(element_ptr n, element_ptr a)
+{
+    fq_data_ptr p = a->data;
+    fq_data_ptr r = n->data;
+    element_ptr nqr = fq_nqr(n->field);
+    element_t e0, e1;
+
+    element_init(e0, p->x->field);
+    element_init(e1, e0->field);
+    element_mul(e0, p->x, p->x);
+    element_mul(e1, p->y, p->y);
+    element_mul(e1, e1, nqr);
+    element_add(e0, e0, e1);
+    element_mul(e1, p->x, p->y);
+    //TODO: which is faster? worth implementing mul_2exp?
+    //element_add(e1, e1, e1);
+    element_mul_si(e1, e1, 2);
+    element_set(r->x, e0);
+    element_set(r->y, e1);
+    element_clear(e0);
+    element_clear(e1);
+}
+
 static void fq_neg(element_ptr n, element_ptr a)
 {
     fq_data_ptr p = a->data;
@@ -153,9 +192,8 @@ static void fq_invert(element_ptr n, element_ptr a)
 
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
-    //TODO: add element_square() because sometimes it's faster
-    element_mul(e0, p->x, p->x);
-    element_mul(e1, p->y, p->y);
+    element_square(e0, p->x);
+    element_square(e1, p->y);
     element_mul(e1, e1, nqr);
     element_sub(e0, e0, e1);
     element_invert(e0, e0);
@@ -216,9 +254,8 @@ static int fq_is_sqr(element_ptr e)
     int result;
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
-    //TODO: element_square() instead:
-    element_mul(e0, p->x, p->x);
-    element_mul(e1, p->y, p->y);
+    element_square(e0, p->x);
+    element_square(e1, p->y);
     element_mul(e1, e1, nqr);
     element_sub(e0, e0, e1);
     result = element_is_sqr(e0);
@@ -241,9 +278,8 @@ static void fq_sqrt(element_ptr n, element_ptr e)
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
     element_init(e2, e0->field);
-    //TODO: element_square() instead:
-    element_mul(e0, p->x, p->x);
-    element_mul(e1, p->y, p->y);
+    element_square(e0, p->x);
+    element_square(e1, p->y);
     element_mul(e1, e1, nqr);
     element_sub(e0, e0, e1);
     element_sqrt(e0, e0);
@@ -287,6 +323,9 @@ void field_init_quadratic(field_ptr f, field_ptr fbase)
     f->sub = fq_sub;
     f->set = fq_set;
     f->mul = fq_mul;
+    f->mul_mpz = fq_mul_mpz;
+    f->mul_si = fq_mul_si;
+    f->square = fq_square;
     f->neg = fq_neg;
     f->cmp = fq_cmp;
     f->invert = fq_invert;
@@ -342,6 +381,27 @@ static void fi_mul(element_ptr n, element_ptr a, element_ptr b)
     element_clear(e2);
 }
 
+static void fi_square(element_ptr n, element_ptr a)
+{
+    fq_data_ptr p = a->data;
+    fq_data_ptr r = n->data;
+    element_t e0, e1;
+
+    element_init(e0, p->x->field);
+    element_init(e1, e0->field);
+    //Re(n) = x^2 - y^2 = (x+y)(x-y)
+    element_add(e0, p->x, p->y);
+    element_sub(e1, p->x, p->y);
+    element_mul(e0, e0, e1);
+    //Im(n) = 2xy
+    element_mul(e1, p->x, p->y);
+    element_add(e1, e1, e1);
+    element_set(r->x, e0);
+    element_set(r->y, e1);
+    element_clear(e0);
+    element_clear(e1);
+}
+
 static void fi_invert(element_ptr n, element_ptr a)
 {
     fq_data_ptr p = a->data;
@@ -350,9 +410,8 @@ static void fi_invert(element_ptr n, element_ptr a)
 
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
-    //TODO: add element_square() because sometimes it's faster
-    element_mul(e0, p->x, p->x);
-    element_mul(e1, p->y, p->y);
+    element_square(e0, p->x);
+    element_square(e1, p->y);
     element_add(e0, e0, e1);
     element_invert(e0, e0);
     element_mul(r->x, p->x, e0);
@@ -371,9 +430,8 @@ static int fi_is_sqr(element_ptr e)
     int result;
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
-    //TODO: element_square() instead:
-    element_mul(e0, p->x, p->x);
-    element_mul(e1, p->y, p->y);
+    element_square(e0, p->x);
+    element_square(e1, p->y);
     element_add(e0, e0, e1);
     result = element_is_sqr(e0);
     element_clear(e0);
@@ -395,9 +453,8 @@ static void fi_sqrt(element_ptr n, element_ptr e)
     element_init(e0, p->x->field);
     element_init(e1, e0->field);
     element_init(e2, e0->field);
-    //TODO: element_square() instead:
-    element_mul(e0, p->x, p->x);
-    element_mul(e1, p->y, p->y);
+    element_square(e0, p->x);
+    element_square(e1, p->y);
     element_add(e0, e0, e1);
     element_sqrt(e0, e0);
     //e0 = sqrt(x^2 + y^2)
@@ -443,6 +500,9 @@ void field_init_fi(field_ptr f, field_ptr fbase)
     f->sub = fq_sub;
     f->set = fq_set;
     f->mul = fi_mul;
+    f->mul_mpz = fq_mul_mpz;
+    f->mul_si = fq_mul_si;
+    f->square = fi_square;
     f->neg = fq_neg;
     f->cmp = fq_cmp;
     f->invert = fi_invert;
