@@ -1,6 +1,6 @@
 #ifndef GET_TIME_H
 #define GET_TIME_H
 
-double get_time();
+double get_time(void);
 
 #endif //GET_TIME_H
