@@ -25,6 +25,10 @@ int main()
 	element_add(z, x, y);
 	element_sub(z, x, y);
     }
+
+    element_clear(x);
+    element_clear(y);
+    element_clear(z);
     mem_report();
     return 0;
 }
